using System;
using System.CodeDom;
using System.Collections.Generic;
using System.IO;
using System.Threading;
using Na.Sql.ParseTree;

namespace Na.Sql
{
    public class EntityBuilder
    {                
        List<string> tableDefinitionPrmaryKey = new List<string>();

        static Type GetType(DataType type)
        {
            switch (type.Type)
            {
                case DataTypes.Char:
                    if (type.Precision != null && type.Precision > 1)
                        return typeof(String);
                    else
                        return typeof(Nullable<Char>);
                case DataTypes.Varchar:
                    return typeof(String);
                case DataTypes.NChar:
                    if (type.Precision != null && type.Precision > 1)
                        return typeof(String);
                    else
                        return typeof(Nullable<Char>);
                case DataTypes.NVarchar:
                    return typeof(String);
                case DataTypes.Int:
                    return typeof(Nullable<Int32>);
                case DataTypes.SmallInt:
                    return typeof(Nullable<Int16>);
                case DataTypes.Decimal:
                    return typeof(Nullable<Decimal>);
                case DataTypes.Real:
                    return typeof(Nullable<Single>);
                case DataTypes.Double:
                    return typeof(Nullable<Double>);
                case DataTypes.Boolean:
                    return typeof(Nullable<Boolean>);
                case DataTypes.Date:
                    return typeof(Nullable<DateTime>);
                case DataTypes.Time:
                    return typeof(Nullable<DateTime>);
                case DataTypes.TimeStamp:
                    return typeof(Nullable<DateTime>);
                default:
                    throw new ArgumentException();
            }
        }

        public CodeCompileUnit Build(string path, string ns)
        {
            Parser parser = new Parser();
            IList<StatementNode> list = parser.Parse(new ParseContext(new Scanner(File.OpenRead(path))));

            CodeNamespace codeNamespace = new CodeNamespace(ns ?? String.Empty);

            foreach (StatementNode node in list)
            {
                tableDefinitionPrmaryKey.Clear();
                CodeTypeDeclaration type = new CodeTypeDeclaration();
                Build(type, node);
                type.CustomAttributes.Add(
                    new CodeAttributeDeclaration(new CodeTypeReference(typeof(SerializableAttribute))));
                type.IsPartial = true;
                codeNamespace.Types.Add(type);
            }

            CodeCompileUnit unit = new CodeCompileUnit();
            unit.Namespaces.Add(codeNamespace);
            
            return unit;
        }
        
        void Build(CodeTypeDeclaration type, StatementNode root)
        {
            Build(type, root.GetStatementPart());
        }

        void Build(CodeTypeDeclaration type, StatementPartNode part)
        {
            Build(type, part.GetCreateStatement());
        }

        void Build(CodeTypeDeclaration type, CreateStatementListNode create)
        {
            Build(type, create.GetTableDefinition());
        }

        void Build(CodeTypeDeclaration type, TableDefinitionNode table)
        {
            type.Name = Thread.CurrentThread.CurrentCulture.TextInfo.ToTitleCase(table.QualifiedName.Name);
            Build(type, table.GetTableElementList());
        }

        void Build(CodeTypeDeclaration type, TableElementListNode elementList)
        {
            foreach (TableConstraintDefinitionNode constraint in elementList.GetTableConstraintDefinitions())
            {
                Build(constraint);
            }

            foreach (ColumnDefinitionNode column in elementList.GetColumnDefinitions())
            {
                Build(type, column);
            }
        }

        void Build(CodeTypeDeclaration type, ColumnDefinitionNode column)
        {
            Type t = GetType(column.GetDataType());
            if (column.NotNull || tableDefinitionPrmaryKey.Contains(column.Name))
                t = Nullable.GetUnderlyingType(t);
            string name = column.Name;

            CodeMemberField field = new CodeMemberField(t, "_" + name.ToLower());
            type.Members.Add(field);

            CodeMemberProperty prop = new CodeMemberProperty();
            prop.HasGet = true;
            prop.HasSet = true;
            prop.Attributes = MemberAttributes.Public | MemberAttributes.Final;
            prop.Name = Thread.CurrentThread.CurrentCulture.TextInfo.ToTitleCase(name);
            prop.Type = new CodeTypeReference(t);

            CodeAssignStatement assign = new CodeAssignStatement(
                new CodeFieldReferenceExpression(new CodeThisReferenceExpression(), field.Name),
                new CodePropertySetValueReferenceExpression());
            prop.SetStatements.Add(assign);

            CodeMethodReturnStatement ret = new CodeMethodReturnStatement(
                new CodeFieldReferenceExpression(new CodeThisReferenceExpression(), field.Name));
            prop.GetStatements.Add(ret);

            type.Members.Add(prop);

        }

        void Build(TableConstraintDefinitionNode constraint)
        {
            if (constraint.IsPrimary)
                tableDefinitionPrmaryKey.AddRange(constraint.Columns);
        }
    }
}
