using System;
using System.Collections.Generic;
using Na.Sql.ParseTree;

namespace Na.Sql
{
    public enum DataTypes
    {
        Char,
        Varchar,
        NChar,
        NVarchar,
        Int,
        SmallInt,
        Decimal,
        Real,
        Double,
        Date,
        Time,
        TimeStamp,
        Boolean,
    }

    public class DataType
    {
        DataTypes type;

        public DataTypes Type
        {
            get { return type; }
        }

        int? precision;
        public int? Precision
        {
            get { return precision; }
        }

        int? scale;
        public int? Scale
        {
            get { return scale; }
        }

        public DataType(DataTypes type) : this(type, null) { }
        public DataType(DataTypes type, int? precision) : this(type, precision, null) { }

        public DataType(DataTypes type, int? precision, int? scale)
        {
            this.type = type;
            this.precision = precision;
            this.scale = scale;
        }
    }

    public class ParseContext
    {
        Scanner scanner = null;
        Token token;
        Token prevToken;

        public ParseContext(Scanner scanner)
        {
            this.scanner = scanner;
            token = null;
            prevToken = null;

            NextToken();
        }

        public void SkipToken(string str)
        {
            if (Match(str))
            {
                NextToken();
            }
            else
            {
                throw new Exception("Expected " + str + ".");
            }
        }

        public void SkipTokenAny(params string[] args)
        {
            if (MatchAny(args))
            {
                NextToken();
            }
            else
            {
                throw new Exception();
            }
        }

        public string Value
        {
            get { return CurrentToken.Str; }
        }

        public int ValueAsInt32
        {
            get { return Int32.Parse(CurrentToken.Str); }
        }

        void NextToken()
        {
            bool lineComment = false;
            bool comment = false;

            if (token != null)
                prevToken = token.Clone() as Token;
            token = scanner.GetToken();
            while (!Match("EOF"))
            {
                if (comment)
                {
                    if (Match("ENDCOMMENT"))
                    {
                        comment = false;
                    }
                    token = scanner.GetToken();
                }
                else if (lineComment)
                {
                    if (Match("CR") || Match("LF"))
                        lineComment = false;
                    else
                        token = scanner.GetToken();
                }
                else
                {
                    if (Match("BEGINCOMMENT"))
                    {
                        comment = true;
                        token = scanner.GetToken();
                    }
                    else if (Match("LINECOMMENT"))
                    {
                        lineComment = true;
                        token = scanner.GetToken();
                    }
                    else if (Match("CR") || Match("LF"))
                    {
                        token = scanner.GetToken();
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }

        public bool Match(string str)
        {
            return str.Equals(token.Def);
        }

        public bool MatchAny(params string[] args)
        {
            foreach (string s in args)
                if (s.Equals(token.Def))
                    return true;
            return false;
        }

        public Token CurrentToken
        {
            get { return token; }
        }

        public Token PreviousToken
        {
            get { return prevToken; }
        }
    }

    public class Parser
    {
        public IList<StatementNode> Parse(ParseContext context)
        {
            List<StatementNode> statementNodeList = new List<StatementNode>();

            while (!context.Match("EOF"))
            {
                if (context.Match("CREATE"))
                {
                    try
                    {
                        StatementNode node = new StatementNode();
                        node.Parse(context);
                        statementNodeList.Add(node);
                    }
                    catch (NotImplementedException)
                    {
                        // do nothing
                    }
                }
                else
                {
                    context.SkipToken(context.CurrentToken.Def);
                }
                   
            }

            return statementNodeList;
        }
    }

    namespace ParseTree
    {
        public interface INode
        {
            void Parse(ParseContext context);
        }

        public class StatementNode : INode
        {
            StatementPartNode statementPartNode = null;

            public StatementPartNode GetStatementPart()
            {
                return statementPartNode;
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                statementPartNode = new StatementPartNode();
                statementPartNode.Parse(context);
                // context.SkipToken("SEMICOLON");
            }

            #endregion
        }

        public class StatementPartNode : INode
        {
            CreateStatementListNode createStatementListNode = null;

            public CreateStatementListNode GetCreateStatement()
            {
                return createStatementListNode;
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                if (context.Match("CREATE"))
                {
                    createStatementListNode = new CreateStatementListNode();
                    createStatementListNode.Parse(context);
                }
                else
                {
                    throw new NotImplementedException();
                }
            }

            #endregion
        }

        public class CreateStatementListNode : INode
        {
            TableDefinitionNode tableDefinitionNode = null;

            public TableDefinitionNode GetTableDefinition()
            {
                return tableDefinitionNode;
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("CREATE");
                if (context.Match("TABLE"))
                {
                    tableDefinitionNode = new TableDefinitionNode();
                    tableDefinitionNode.Parse(context);
                }
                else
                {
                    // create tableȊO͖
                    throw new NotImplementedException();
                }
            }

            #endregion
        }

        public class TableDefinitionNode : INode
        {
            QualifiedNameNode qualifiedNameNode = null;
            TableElementListNode tableElementListNode = null;

            public QualifiedNameNode QualifiedName
            {
                get { return qualifiedNameNode; }
            }

            public TableElementListNode GetTableElementList()
            {
                return tableElementListNode;
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("TABLE");

                qualifiedNameNode = new QualifiedNameNode();
                qualifiedNameNode.Parse(context);

                tableElementListNode = new TableElementListNode();
                tableElementListNode.Parse(context);
            }

            #endregion
        }

        public class QualifiedNameNode : INode
        {
            List<string> identifiers = null;

            public IList<string> FullName
            {
                get { return identifiers; }
            }

            public string Name
            {
                get
                {
                    return identifiers[identifiers.Count - 1];
                }
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                identifiers = new List<string>();
                identifiers.Add(context.CurrentToken.Str);

                context.SkipToken("IDENT");
                if (context.Match("PERIOD"))
                {
                    context.SkipToken("PERIOD");
                    identifiers.Add(context.CurrentToken.Str);
                    context.SkipToken("IDENT");
                }
            }

            #endregion
        }

        public class TableElementListNode : INode
        {
            List<TableElementNode> tableElements = null;

            public IList<ColumnDefinitionNode> GetColumnDefinitions()
            {
                List<ColumnDefinitionNode> list = new List<ColumnDefinitionNode>();
                foreach (TableElementNode element in tableElements)
                {
                    ColumnDefinitionNode column = element.GetChild() as ColumnDefinitionNode;
                    if (column != null)
                        list.Add(column);
                }
                return list;
            }

            public IList<TableConstraintDefinitionNode> GetTableConstraintDefinitions()
            {
                List<TableConstraintDefinitionNode> list = new List<TableConstraintDefinitionNode>();
                foreach (TableElementNode element in tableElements)
                {
                    TableConstraintDefinitionNode constraint = element.GetChild() as TableConstraintDefinitionNode;
                    if (constraint != null)
                        list.Add(constraint);
                }
                return list;
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("OPEN");

                tableElements = new List<TableElementNode>();
                TableElementNode node = new TableElementNode();
                node.Parse(context);
                tableElements.Add(node);
                while (context.Match("COMMA"))
                {
                    context.SkipToken("COMMA");
                    node = new TableElementNode();
                    node.Parse(context);
                    tableElements.Add(node);
                }
                context.SkipToken("CLOSE");
            }

            #endregion
        }

        class TableElementNode : INode
        {
            INode element = null;

            public INode GetChild()
            {
                return element;
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                if (context.Match("IDENT"))
                    element = new ColumnDefinitionNode();
                else
                    element = new TableConstraintDefinitionNode();

                element.Parse(context);
            }

            #endregion
        }

        public class ColumnDefinitionNode : INode
        {
            string identifier = null;
            DataTypeNode dataTypeNode = null;
            DefaultAndConstraintListNode defaultAndConstraintListNode = null;

            public string Name
            {
                get { return identifier; }
            }

            public DataType GetDataType()
            {
                return dataTypeNode.GetDataType();
            }

            public bool NotNull
            {
                get
                {
                    return defaultAndConstraintListNode == null ?
                        false : defaultAndConstraintListNode.NotNull;
                }
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                identifier = context.Value;
                context.SkipToken("IDENT");

                dataTypeNode = new DataTypeNode();
                dataTypeNode.Parse(context);

                switch (context.CurrentToken.Def)
                {
                    case "CONSTRAINT": goto case "<DefaultAndConstraint>";
                    case "NOT": goto case "<DefaultAndConstraint>";
                    case "NULL": goto case "<DefaultAndConstraint>";
                    case "UNIQUE": goto case "<DefaultAndConstraint>";
                    case "PRIMARY": goto case "<DefaultAndConstraint>";
                    case "REFERENCES": goto case "<DefaultAndConstraint>";
                    case "CHECK": goto case "<DefaultAndConstraint>";
                    case "DEFAULT": goto case "<DefaultAndConstraint>";
                    case "<DefaultAndConstraint>":
                        defaultAndConstraintListNode = new DefaultAndConstraintListNode();
                        defaultAndConstraintListNode.Parse(context);
                        break;
                    default:
                        break;
                }
            }

            #endregion
        }

        public class TableConstraintDefinitionNode : INode
        {
            ConstraintNameDefinitionNode constraintNameDefinitionNode = null;
            TableConstraintNode tableConstraintNode = null;

            public bool IsPrimary
            {
                get { return tableConstraintNode.IsPrimary; }
            }

            public IList<string> Columns
            {
                get { return tableConstraintNode.Columns; }
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                if (context.Match("CONSTRAINT"))
                {
                    constraintNameDefinitionNode = new ConstraintNameDefinitionNode();
                    constraintNameDefinitionNode.Parse(context);
                }

                tableConstraintNode = new TableConstraintNode();
                tableConstraintNode.Parse(context);
            }

            #endregion
        }

        class TableConstraintNode : INode
        {
            UniqueConstraintDefinitionNode primaryConstraintDefinitionNode = null;
            UniqueConstraintDefinitionNode uniqueConstraintDefinitionNode = null;
            ReferentialConstraintDefinitionNode referentialConstraintDefinition = null;
            IList<string> columns;

            public bool IsPrimary
            {
                get { return primaryConstraintDefinitionNode != null; }
            }

            public IList<string> Columns
            {
                get { return columns; }
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.Match("PRIMARY"))
                {
                    primaryConstraintDefinitionNode = new UniqueConstraintDefinitionNode();
                    primaryConstraintDefinitionNode.Parse(context);
                    columns = primaryConstraintDefinitionNode.Columns;
                }
                else if (context.Match("UNIQUE"))
                {
                    uniqueConstraintDefinitionNode = new UniqueConstraintDefinitionNode();
                    uniqueConstraintDefinitionNode.Parse(context);
                    columns = uniqueConstraintDefinitionNode.Columns;
                }
                else if (context.Match("FOREIGN"))
                {
                    referentialConstraintDefinition = new ReferentialConstraintDefinitionNode();
                    referentialConstraintDefinition.Parse(context);
                }
                else if (context.Match("CHECK"))
                {
                    throw new NotImplementedException();
                }
            }

            #endregion
        }

        class UniqueConstraintDefinitionNode : INode
        {
            UniqueSpecificationNode uniqueSpecficationNode = null;
            UniqueColumnListNode uniqueColumnListNode = null;

            public IList<string> Columns
            {
                get { return uniqueColumnListNode.Columns; }
            }
            #region INode o

            public void Parse(ParseContext context)
            {
                uniqueSpecficationNode = new UniqueSpecificationNode();
                uniqueSpecficationNode.Parse(context);

                context.SkipToken("OPEN");
                uniqueColumnListNode = new UniqueColumnListNode();
                uniqueColumnListNode.Parse(context);
                context.SkipToken("CLOSE");
            }

            #endregion
        }

        class UniqueColumnListNode : INode
        {
            ColumnNameListNode columnNameListNode = null;

            public IList<string> Columns
            {
                get
                {
                    return columnNameListNode.Names;
                }
            }
            #region INode o

            public void Parse(ParseContext context)
            {
                columnNameListNode = new ColumnNameListNode();
                columnNameListNode.Parse(context);
            }

            #endregion
        }

        class ReferentialConstraintDefinitionNode : INode
        {
            ColumnNameListNode columnNameListNode = null;
            ReferencesSpecificationNode referencesSpecificationNode = null;

            #region INode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("FOREIGN");
                context.SkipToken("KEY");
                context.SkipToken("OPEN");
                columnNameListNode = new ColumnNameListNode();
                columnNameListNode.Parse(context);
                context.SkipToken("CLOSE");
                referencesSpecificationNode = new ReferencesSpecificationNode();
                referencesSpecificationNode.Parse(context);
            }

            #endregion
        }

        interface IDataTypeNode : INode
        {
            DataType GetDataType();
        }

        class DataTypeNode : IDataTypeNode
        {
            DataTypeCommonNode dataTypeCommonNode = null;

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                dataTypeCommonNode = new DataTypeCommonNode();
                dataTypeCommonNode.Parse(context);
            }

            #endregion

            #region IDataTypeNode o

            public DataType GetDataType()
            {
                return dataTypeCommonNode.GetDataType();
            }

            #endregion
        }

        class DataTypeCommonNode : IDataTypeNode
        {
            DataTypes type;
            IDataTypeNode typeNode = null;

            public DataType GetDataType()
            {
                if (typeNode != null)
                {
                    return typeNode.GetDataType();
                }
                else
                {
                    return new DataType(type);
                }
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                switch (context.CurrentToken.Def)
                {
                    case "CHAR": goto case "<CharacterString>";
                    case "CHARACTER": goto case "<CharacterString>";
                    case "VARCHAR": goto case "<CharacterString>";
                    case "VARCHAR2": // for Oracle
                        context.CurrentToken.Def = "VARCHAR";
                        goto case "<CharacterString>";
                    case "NATIONAL": goto case "<NationalCharacterString>";
                    case "NCHAR": goto case "<NationalCharacterString>";
                    case "NVARCHAR": goto case "<NationalCharacterString>";
                    case "INTEGER": goto case "<Numeric>";
                    case "INT": goto case "<Numeric>";
                    case "SMALLINT": goto case "<Numeric>";
                    case "NUMERIC": goto case "<Numeric>";
                    case "DECIMAL": goto case "<Numeric>";
                    case "NUMBER": // for Oracle
                        context.CurrentToken.Def = "NUMERIC";
                        goto case "<Numeric>";
                    case "DEC": goto case "<Numeric>";
                    case "FLOAT": goto case "<Numeric>";
                    case "REAL": goto case "<Numeric>";
                    case "DOUBLE": goto case "<Numeric>";
                    case "DATE": goto case "<DateTime>";
                    case "TIME": goto case "<DateTime>";
                    case "TIMESTAMP": goto case "<DateTime>";
                    case "BOOLEAN":
                        type = DataTypes.Boolean;
                        context.SkipToken("BOOLEAN");
                        break;
                    case "<CharacterString>":
                        typeNode = new CharacterStringTypeNode();
                        typeNode.Parse(context);
                        break;
                    case "<NationalCharacterString>":
                        typeNode = new NationalCharacterStringTypeNode();
                        typeNode.Parse(context);
                        break;
                    case "<Numeric>":
                        typeNode = new NumericTypeNode();
                        typeNode.Parse(context);
                        break;
                    case "<DateTime>":
                        typeNode = new DateTimeTypeNode();
                        typeNode.Parse(context);
                        break;
                    default:
                        throw new NotImplementedException();
                }
            }

            #endregion
        }

        class CharacterStringTypeNode : IDataTypeNode
        {
            DataTypes type;
            CharLengthNode charLengthNode = null;

            public DataType GetDataType()
            {
                return new DataType(type, charLengthNode == null ? null : (int?)charLengthNode.Value);
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.Match("VARCHAR"))
                {
                    type = DataTypes.Varchar;
                    context.SkipToken("VARCHAR");
                    charLengthNode = new CharLengthNode();
                    charLengthNode.Parse(context);
                }
                else if (context.MatchAny("CHAR", "CHARACTER"))
                {
                    CharOrCharacterNode node = new CharOrCharacterNode();
                    node.Parse(context);
                    type = node.GetDataType().Type;

                    if (context.Match("VARYING"))
                    {
                        type = DataTypes.Varchar;
                        context.SkipToken("VARYING");
                        charLengthNode = new CharLengthNode();
                        charLengthNode.Parse(context);
                    }
                    else if (context.Match("OPEN"))
                    {
                        charLengthNode = new CharLengthNode();
                        charLengthNode.Parse(context);
                    }
                }
            }

            #endregion
        }

        class CharLengthNode : INode
        {
            int value;

            public int Value
            {
                get { return value; }
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("OPEN");
                value = context.ValueAsInt32;
                context.SkipToken("NUM");
                if (context.MatchAny("BYTE", "CHAR"))
                {
                    context.SkipTokenAny("BYTE", "CHAR");
                }
                context.SkipToken("CLOSE");
            }

            #endregion
        }

        class CharOrCharacterNode : IDataTypeNode
        {
            DataTypes type;

            public DataType GetDataType()
            {
                return new DataType(type);
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                type = DataTypes.Char;
                context.SkipTokenAny("CHAR", "CHARACTER");
            }

            #endregion
        }

        class NationalCharacterStringTypeNode : IDataTypeNode
        {
            DataTypes type;
            CharLengthNode charLengthNode = null;

            public DataType GetDataType()
            {
                return new DataType(type, charLengthNode == null ? null : (int?)charLengthNode.Value);
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.Match("NATIONAL"))
                {
                    context.SkipToken("NATIONAL");
                    CharOrCharacterNode node = new CharOrCharacterNode();
                    node.Parse(context);
                    type = node.GetDataType().Type;

                    if (context.Match("VARYING"))
                    {
                        type = DataTypes.NVarchar;
                        context.SkipToken("VARYING");
                        charLengthNode = new CharLengthNode();
                        charLengthNode.Parse(context);
                    }
                    else if (context.Match("OPEN"))
                    {
                        charLengthNode = new CharLengthNode();
                        charLengthNode.Parse(context);
                    }
                }
                else if (context.Match("NCHAR"))
                {
                    type = DataTypes.NChar;
                    context.SkipToken("NCHAR");
                    if (context.Match("VARYING"))
                    {
                        type = DataTypes.NVarchar;
                        context.SkipToken("VARYING");
                        charLengthNode = new CharLengthNode();
                        charLengthNode.Parse(context);
                    }
                    else if (context.Match("OPEN"))
                    {
                        charLengthNode = new CharLengthNode();
                        charLengthNode.Parse(context);
                    }
                }
                else if (context.Match("NVARCHAR"))
                {
                    type = DataTypes.NVarchar;
                    charLengthNode = new CharLengthNode();
                    charLengthNode.Parse(context);
                }
            }

            #endregion
        }

        class NumericTypeNode : IDataTypeNode
        {
            IDataTypeNode numericNode = null;

            public DataType GetDataType()
            {
                return numericNode.GetDataType();
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                switch (context.CurrentToken.Def)
                {
                    case "INTEGER": goto case "<ExactNumeric>";
                    case "INT": goto case "<ExactNumeric>";
                    case "SMALLINT": goto case "<ExactNumeric>";
                    case "NUMERIC": goto case "<ExactNumeric>";
                    case "DECIMAL": goto case "<ExactNumeric>";
                    case "DEC": goto case "<ExactNumeric>";
                    case "REAL": goto case "<ApproximateNumeric>";
                    case "FLOAT": goto case "<ApproximateNumeric>";
                    case "DOUBLE": goto case "<ApproximateNumeric>";
                    case "<ExactNumeric>":
                        numericNode = new ExactNumericTypeNode();
                        break;
                    case "<ApproximateNumeric>":
                        numericNode = new ApproximateNumericType();
                        break;
                    default:
                        throw new NotImplementedException();
                }
                numericNode.Parse(context);
            }

            #endregion
        }

        class ExactNumericTypeNode : IDataTypeNode
        {
            DataTypes type;
            PrecisionNode precisionNode = null;
            ScaleNode scaleNode = null;
            ExactIntegerTypeNode exactIntegerTypeNode = null;

            public DataType GetDataType()
            {
                if (exactIntegerTypeNode != null)
                    return exactIntegerTypeNode.GetDataType();
                else
                    return new DataType(type, precisionNode == null ? null : (int?)precisionNode.Value,
                        scaleNode == null ? null : (int?)scaleNode.Value);
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.MatchAny("NUMERIC", "DECIMAL", "DEC"))
                {
                    type = DataTypes.Decimal;
                    context.SkipTokenAny("NUMERIC", "DECIMAL", "DEC");
                    if (context.Match("OPEN"))
                    {
                        context.SkipToken("OPEN");
                        precisionNode = new PrecisionNode();
                        precisionNode.Parse(context);
                        if (context.Match("COMMA"))
                        {
                            context.SkipToken("COMMA");
                            scaleNode = new ScaleNode();
                            scaleNode.Parse(context);
                        }
                        context.SkipToken("CLOSE");
                    }
                }
                else
                {
                    exactIntegerTypeNode = new ExactIntegerTypeNode();
                    exactIntegerTypeNode.Parse(context);
                }
            }

            #endregion
        }

        class ApproximateNumericType : IDataTypeNode
        {
            DataTypes type;
            PrecisionNode precisionNode = null;

            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.Match("FLOAT"))
                {
                    type = DataTypes.Double;
                    context.SkipToken("FLOAT");
                    precisionNode = new PrecisionNode();
                    precisionNode.Parse(context);
                }
                else if (context.Match("REAL"))
                {
                    type = DataTypes.Real;
                    context.SkipToken("REAL");
                }
                else
                {
                    DoubleNode node = new DoubleNode();
                    node.Parse(context);
                    type = node.GetDataType().Type;
                }
            }

            #endregion

            #region IDataTypeNode o

            public DataType GetDataType()
            {
                return new DataType(type, precisionNode.Value);
            }

            #endregion
        }

        class DoubleNode : IDataTypeNode
        {
            DataTypes type;

            #region INode o

            public void Parse(ParseContext context)
            {
                type = DataTypes.Double;
                context.SkipToken("DOUBLE");
                if (context.Match("PRECISION"))
                {
                    context.SkipToken("PRECISION");
                }
            }

            #endregion

            #region IDataTypeNode o

            public DataType GetDataType()
            {
                return new DataType(type);
            }

            #endregion
        }


        class PrecisionNode : INode
        {
            int value;

            public int Value
            {
                get { return value; }
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                value = context.ValueAsInt32;
                context.SkipToken("NUM");
            }

            #endregion
        }

        class ScaleNode : INode
        {
            int value;

            public int Value
            {
                get { return value; }
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                value = context.ValueAsInt32;
                context.SkipToken("NUM");
            }

            #endregion
        }

        class ExactIntegerTypeNode : IDataTypeNode
        {
            DataTypes type;

            public DataType GetDataType()
            {
                return new DataType(type);
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.Match("SMALLINT"))
                    type = DataTypes.SmallInt;
                else
                    type = DataTypes.Int;
                context.SkipTokenAny("INTEGER", "INT", "SMALLINT");
            }

            #endregion
        }


        class DateTimeTypeNode : IDataTypeNode
        {
            DataTypes type;

            public DataType GetDataType()
            {
                return new DataType(type);
            }

            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.Match("DATE"))
                    type = DataTypes.Date;
                else if (context.Match("TIME"))
                    type = DataTypes.Time;
                else
                    type = DataTypes.TimeStamp;

                context.SkipTokenAny("DATE", "TIME", "TIMESTAMP");
            }

            #endregion
        }

        class DefaultAndConstraintListNode : INode
        {
            DefaultClauseNode defaultClauseNode = null;
            List<ColumnConstraintDefinitionNode> columnConstraintDefinitionNodeList = null;

            public bool NotNull
            {
                get
                {
                    foreach (ColumnConstraintDefinitionNode node in columnConstraintDefinitionNodeList)
                        if (node.NotNull.HasValue && node.NotNull.Value)
                            return true;
                    return false;
                }
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                columnConstraintDefinitionNodeList = new List<ColumnConstraintDefinitionNode>();

                ColumnConstraintDefinitionNode node = null;
                bool foundDefault = false;
                bool loop = true;

                while (loop)
                {
                    switch (context.CurrentToken.Def)
                    {
                        case "DEFAULT":
                            if (foundDefault)
                                throw new Exception("Duplicate default.");

                            foundDefault = true;
                            defaultClauseNode = new DefaultClauseNode();
                            defaultClauseNode.Parse(context);
                            break;
                        case "CONSTRAINT": goto case "<ColumnConstraintDefinition>";
                        case "NOT": goto case "<ColumnConstraintDefinition>";
                        case "NULL": goto case "<ColumnConstraintDefinition>";
                        case "UNIQUE": goto case "<ColumnConstraintDefinition>";
                        case "PRIMARY": goto case "<ColumnConstraintDefinition>";
                        case "REFERENCES": goto case "<ColumnConstraintDefinition>";
                        case "CHECK": goto case "<ColumnConstraintDefinition>";
                        case "<ColumnConstraintDefinition>":
                            node = new ColumnConstraintDefinitionNode();
                            node.Parse(context);
                            columnConstraintDefinitionNodeList.Add(node);
                            break;
                        default:
                            loop = false;
                            break;
                    }
                }
            }

            #endregion
        }

        class DefaultClauseNode : INode
        {
            DefaultOptionNode defaultOptionNode = null;

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("DEFAULT");
                if (context.Match("GENERATED"))
                {
                    throw new NotImplementedException();
                }
                else
                {
                    defaultOptionNode = new DefaultOptionNode();
                    defaultOptionNode.Parse(context);
                }
            }

            #endregion
        }

        class DefaultOptionNode : INode
        {
            string value;
            LiteralNode literalNode = null;
            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.Match("NULL"))
                {
                    value = "NULL";
                    context.SkipToken("NULL");
                }
                else
                {
                    literalNode = new LiteralNode();
                    literalNode.Parse(context);
                }
            }

            #endregion
        }

        class LiteralNode : INode
        {
            string value;
            SignNode signNode = null;
            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.MatchAny("PLUS", "MINUS", "NUM", "DBL"))
                {
                    if (context.MatchAny("PLUS", "MINUS"))
                    {
                        signNode = new SignNode();
                        signNode.Parse(context);
                    }
                    value = context.Value;
                    context.SkipTokenAny("NUM", "DBL");
                }
                else if (context.Match("STR"))
                {
                    value = context.Value;
                    context.SkipToken("STR");
                }
            }

            #endregion
        }

        class SignNode : INode
        {
            bool negative;
            #region INode o

            public void Parse(ParseContext context)
            {
                negative = context.Value == "MINU";
                context.SkipTokenAny("PLUS", "MINUS");
            }

            #endregion
        }


        class ColumnConstraintDefinitionNode : INode
        {
            ConstraintNameDefinitionNode constraintNameDefinitionNode = null;
            ColumnConstraintNode columnConstraintNode = null;

            public bool? NotNull
            {
                get { return columnConstraintNode.NotNull; }
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                if (context.Match("CONSTRAINT"))
                {
                    constraintNameDefinitionNode = new ConstraintNameDefinitionNode();
                    constraintNameDefinitionNode.Parse(context);
                }

                columnConstraintNode = new ColumnConstraintNode();
                columnConstraintNode.Parse(context);
            }

            #endregion
        }

        class ConstraintNameDefinitionNode : INode
        {
            QualifiedNameNode qualifiedNameNode = null;

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("CONSTRAINT");
                qualifiedNameNode = new QualifiedNameNode();
                qualifiedNameNode.Parse(context);
            }

            #endregion
        }

        class ColumnConstraintNode : INode
        {
            bool? notNull = null;
            INode constraint = null;

            public bool? NotNull
            {
                get { return notNull; }
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                switch (context.CurrentToken.Def)
                {
                    case "NOT":
                        context.SkipToken("NOT");
                        context.SkipToken("NULL");
                        notNull = true;
                        break;
                    case "NULL":
                        context.SkipToken("NULL");
                        notNull = false;
                        break;
                    case "PRIMARY":
                        notNull = true;
                        goto case "UNIQUE";
                    case "UNIQUE":
                        constraint = new UniqueSpecificationNode();
                        constraint.Parse(context);
                        break;
                    case "REFERENCES":
                        constraint = new ReferencesSpecificationNode();
                        constraint.Parse(context);
                        break;
                    case "CHECK":
                        constraint = new CheckConstraintDefinitionNode();
                        constraint.Parse(context);
                        break;
                    default:
                        throw new Exception();
                }
            }

            #endregion
        }

        class UniqueSpecificationNode : INode
        {
            bool primary = false;
            bool unique = false;

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                if (context.Match("PRIMARY"))
                {
                    context.SkipToken("PRIMARY");
                    context.SkipToken("KEY");
                    primary = true;
                }
                else
                {
                    context.SkipToken("UNIQUE");
                    unique = true;
                }
            }

            #endregion
        }

        class ReferencesSpecificationNode : INode
        {
            ReferencedTableAndColumnListNode referencedTableAndColumnListNode = null;

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("REFERENCES");
                referencedTableAndColumnListNode = new ReferencedTableAndColumnListNode();
                referencedTableAndColumnListNode.Parse(context);
            }

            #endregion
        }

        class CheckConstraintDefinitionNode : INode
        {
            ValueExpressionNode valueExpressionNode = null;
            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("CHECK");
                context.SkipToken("OPEN");
                valueExpressionNode = new ValueExpressionNode();
                valueExpressionNode.Parse(context);
                context.SkipToken("CLOSE");
            }

            #endregion
        }

        class ReferencedTableAndColumnListNode : INode
        {
            QualifiedNameNode qualifiedNameNode = null;
            ColumnNameListNode columnNameListNode = null;


            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                qualifiedNameNode = new QualifiedNameNode();
                qualifiedNameNode.Parse(context);

                if (context.Match("OPEN"))
                {
                    context.SkipToken("OPEN");
                    columnNameListNode = new ColumnNameListNode();
                    columnNameListNode.Parse(context);
                    context.SkipToken("CLOSE");
                }
            }

            #endregion
        }

        class ColumnNameListNode : INode
        {
            List<string> names = null;

            public IList<string> Names
            {
                get { return names; }
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                names = new List<string>();
                ColumnNameItem name = new ColumnNameItem();
                name.Parse(context);
                names.Add(name.Value);
                while (context.Match("COMMA"))
                {
                    context.SkipToken("COMMA");
                    name = new ColumnNameItem();
                    name.Parse(context);
                    names.Add(name.Value);
                }
            }

            #endregion
        }

        class ColumnNameItem : INode
        {
            string identifier = null;

            public string Value
            {
                get { return identifier; }
            }

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                identifier = context.Value;
                context.SkipToken("IDENT");
            }

            #endregion
        }

        class ValueExpressionNode : INode
        {
            List<OrExpressionNode> expressions = null;

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                OrExpressionNode expression = new OrExpressionNode();
                expression.Parse(context);
                expressions.Add(expression);
                while (context.Match("OR"))
                {
                    context.SkipToken("OR");
                    expression = new OrExpressionNode();
                    expression.Parse(context);
                    expressions.Add(expression);
                }
            }

            #endregion
        }

        class OrExpressionNode : INode
        {
            List<AndExpressionNode> expressions = null;

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                AndExpressionNode expression = new AndExpressionNode();
                expression.Parse(context);
                expressions.Add(expression);
                while (context.Match("AND"))
                {
                    context.SkipToken("AND");
                    expression = new AndExpressionNode();
                    expression.Parse(context);
                    expressions.Add(expression);
                }
            }

            #endregion
        }

        class AndExpressionNode : INode
        {
            IsSearchConditionNode isSearchConditionNode = null;
            bool isNot = false;

            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                if (context.Match("NOT"))
                {
                    isNot = true;
                }

                isSearchConditionNode = new IsSearchConditionNode();
                isSearchConditionNode.Parse(context);
            }

            #endregion
        }

        class IsSearchConditionNode : INode
        {
            BooleanPrimaryNode booleanPrimaryNode = null;
            #region ISqlNode o

            public void Parse(ParseContext context)
            {
                booleanPrimaryNode = new BooleanPrimaryNode();
                booleanPrimaryNode.Parse(context);
                if (context.Match("IS"))
                {
                    context.SkipToken("IS");
                    if (context.Match("NOT"))
                    {
                        context.SkipToken("NOT");
                    }
                    context.SkipToken("NULL");
                }
            }

            #endregion
        }

        class BooleanPrimaryNode : INode
        {
            PredicateNode predicate = null;

            #region INode o

            public void Parse(ParseContext context)
            {
                predicate.Parse(context);
            }

            #endregion
        }

        class PredicateNode : INode
        {
            ExistsExpressionNode existsExpressionNode = null;
            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.Match("EXISTS"))
                {
                    existsExpressionNode = new ExistsExpressionNode();
                    existsExpressionNode.Parse(context);
                }

            }

            #endregion
        }

        class AdditiveExpressionNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }

        class ExistsExpressionNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("EXISTS");
                context.SkipToken("OPEN");
                context.SkipToken("CLOSE");
            }

            #endregion
        }

        class RemainingPredicateNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }

        class TableSubQueryNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }

        class SubQueryNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }

        class QueryExpressionNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }

        class NonJoinQueryTermNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }

        class NonJoinQueryPrimaryNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }

        class SimpleTableNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }

        class TableValueConstructorNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }

        class QuerySpecificationNode : INode
        {
            SetQuantifierNode setQuantifierNode = null;

            #region INode o

            public void Parse(ParseContext context)
            {
                context.SkipToken("SELECT");
                if (context.Match("DISTINCT") || context.Match("ALL"))
                {
                    setQuantifierNode = new SetQuantifierNode();
                    setQuantifierNode.Parse(context);
                }

                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }

        class SetQuantifierNode : INode
        {
            bool isDistinct = false;

            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.Match("DISTINCT"))
                {
                    isDistinct = true;
                    context.SkipToken("DISTINCT");
                }
                else if (context.Match("ALL"))
                {
                    context.SkipToken("ALL");
                }
            }

            #endregion
        }

        class SelectListNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                if (context.Match("ASTERISK"))
                {
                    context.SkipToken("ASTERISK");
                }
                else
                {
                }
            }

            #endregion
        }

        class SelectColumnListNode : INode
        {
            List<SelectSublistNode> selectSublist = null;

            #region INode o

            public void Parse(ParseContext context)
            {
                selectSublist = new List<SelectSublistNode>();

                SelectSublistNode node = new SelectSublistNode();
                node.Parse(context);
                selectSublist.Add(node);

                while (context.Match("COMMA"))
                {
                    context.SkipToken("COMMA");
                    node = new SelectSublistNode();
                    node.Parse(context);
                    selectSublist.Add(node);
                }
            }

            #endregion
        }

        class SelectSublistNode : INode
        {
            #region INode o

            public void Parse(ParseContext context)
            {
                throw new Exception("The method or operation is not implemented.");
            }

            #endregion
        }
    }
}